mod.service("regressionFunc", [
    function () {

        //	Define allowed chart types
        let supportedChartTypes =
            ["chart/column", "chart/line", "chart/bar", "chart/area", "chart/scatter", "trelliswidget"];

        let epsilon = 0.000001,
            regressionErrorKey = "regressionError",
            errors = {
            exponential: {
                zeros: "Data contains zeros. Cannot calculate regression.",
                negative: "Cannot calculate exponential regression on negative values",
                nulls: "Cannot calculate exponential regression on null values"
            },
            logarithmic:{
                zeros: "Data contains zeros. Cannot calculate regression.",
                negative: "Cannot calculate exponential regression on negative values"
            },
            defaultError: "Cannot calculate regression line"
        };


        //	Function to kick off a regression analysis
        this.startRegression = (widget, args) => {

            // Add hook for the quadrant analysis
            if (!this.widgetIsSupported(widget.type)) {
                return;
            }

            let valuePanelName = "values",
                isTrellis = widget.type === "trelliswidget",
                categoryPanelIndex = 0;

            //	Get a list of the value items and categories
            let valueItems = this.getItems(widget, valuePanelName),
                categories = this.getItems(widget, categoryPanelIndex);

            if ($$get(valueItems, "length") > 0 && $$get(categories, "length") > 0) {

                let category = categories[0];

                if (isTrellis) {
                    let data = args.result.data;

                    // Trellis
                    for (let key in data) {
                        if (data.hasOwnProperty(key)) {
                            let chart = data[key].chart;
                            this.setRegressionOnCharts(widget, chart, valueItems, category);
                        }
                    }
                } else {
                    let chart = args.result;
                    this.setRegressionOnCharts(widget, chart, valueItems, category);
                }
            }
        };


        this.setRegressionOnCharts = (widget, chart, valueItems, category) => {
            let additionalSeries = [];
            valueItems.forEach((item, i)=>{
                let series = chart.series[i];

                //	Should regression be calculated for this series?
                let regressionEnabled = (item[mod.regressionProperty]) &&
                    (item[mod.regressionProperty] !== mod.none);

                if (regressionEnabled) {

                    //	Get saved settings
                    let settings = widget.options.regression || $.extend(true, {}, mod.defaultSettings),
                        duplicatedSeries = $.extend(true, {}, series);

                    series.regression = false;

                    //	Enable regression for this series
                    duplicatedSeries.regression = true;

                    //	Define the future periods to predict out
                    let predictionOptions = definePredictionOptions(duplicatedSeries, category, settings);

                    let color = widget.type === "trelliswidget" ?
                        duplicatedSeries.color : item.format.color.color;

                    //	Set the options for this series
                    duplicatedSeries.regressionSettings = {
                        type: $$get(duplicatedSeries,"regressionSettings.type") ||
                            item[mod.regressionProperty] || "linear",
                        color: $$get(duplicatedSeries,"regressionSettings.color") ||  color,
                        dashStyle: $$get(duplicatedSeries, "regressionSettings.dashStyle") || "dash",
                        name: $$get(duplicatedSeries, "regressionSettings.name") || item.jaql.title + " (Regression)",
                        projections: predictionOptions,
                        style: widget.style,
                        hideInLegend:  $$get(duplicatedSeries, "regressionSettings.hideInLegend") || false,
                        lineWidth:  $$get(duplicatedSeries, "regressionSettings.lineWidth") || 2,
                        tooltip: $$get(duplicatedSeries, "regressionSettings.tooltip") || {},
                        decimalPlaces:  $$get(duplicatedSeries, "regressionSettings.decimalPlaces") || 2,
                        useAllSeries: $$get(duplicatedSeries, "regressionSettings.useAllSeries") || false,
                        lineType: $$get(duplicatedSeries, "regressionSettings.lineType") || "spline",
                    };

                    Object.assign(duplicatedSeries, _.pick(duplicatedSeries.regressionSettings,
                        "name",
                        "lineWidth",
                        "dashStyle",
                        "color"));

                    //duplicatedSeries.name =  duplicatedSeries.regressionSettings.name;
                    //duplicatedSeries.lineWidth = duplicatedSeries.regressionSettings.lineWidth;
                    duplicatedSeries.marker =  {enabled: false};
                    duplicatedSeries.isRegressionLine = true;
                    duplicatedSeries.type = duplicatedSeries.regressionSettings.lineType;
                    //duplicatedSeries.dashStyle = duplicatedSeries.regressionSettings.dashStyle;
                    duplicatedSeries.showInLegend = !duplicatedSeries.regressionSettings.hideInLegend;
                    //duplicatedSeries.color = duplicatedSeries.regressionSettings.color;

                    // Update the chart to show future points
                    addFuturePeriodsAxisLabels(widget, chart, predictionOptions, i);

                    additionalSeries.push(duplicatedSeries);

                    // Calculate trendline, and replace the old series
                    chart.series.push(duplicatedSeries);

                    calcRegression(duplicatedSeries, chart);
                }
            });
        };

        //	function to find the value items from a widget
        this.getItems = (widget, panelNameOrNumer) => {

            let panel;

            if (!isNaN(panelNameOrNumer)){
                panel = widget.metadata.panels[panelNameOrNumer];
            } else if (!defined(panelNameOrNumer)){
                panel = widget.metadata.panels[0];
            } else {
                panel = widget.metadata.panel(panelNameOrNumer);
            }

            //	Get only the enabled items
            return panel.getEnabledItems();
        };

        //	Function to determine if the regression option should be added
        this.canEnableRegression = function(e, args, headerMenuCaption, menuType){
            try {
                //	Has the menu been added already?
                let item  = _.find(args.settings.items, (item)=>{
                    return item.caption === headerMenuCaption;
                });

                if (item){
                    return false;
                }

                // True for values metadata panel
                if (menuType === "measure" &&
                    ($$get(args, "settings.name") === "widget-metadataitem" &&
                    $$get(args,"settings.item.$$panel.name") === "values")){
                    return true;
                }
                // For widget menu on edit mode
                else if ((menuType === "settings" &&
                        args.ui.css === "w-menu-host")){
                    return true;
                }
            } catch(ex) {
                return false;
            }

            //	Return result
            return false;
        };


        //	Function to determine if this chart type is supported
        this.widgetIsSupported = (type) => {
            return supportedChartTypes.indexOf(type) >= 0;
        };

        //	Figure out how many periods to predict out
        function definePredictionOptions(series, dim, settings){

            //	Create options object
            let options = {
                enabled: false,
                predictionMethod: settings.predictionMethod
            };

            //	Is the axis a datetime?
            if (dim.jaql.datatype !== "datetime" || settings.predictionMethod === mod.none) {

                //	No need to predict into the future, as the axis is not a datetime
                return options;

            } else {

                //	Need to predict into the future
                options.enabled = true;

                //	Get the formatting options
                options.level = dim.jaql.level;
                options.mask = dim.format.mask[dim.jaql.level];

                //	Figure out how far to predict ahead
                let maxDate = series.data[series.data.length-1].selectionData[0];
                options.projectedPeriods =  calculateFuturePeriods(maxDate,options);
            }

            return options;
        }

        //  Function to determine the # of future periods
        function calculateFuturePeriods(maxDate, options) {

            //	Get parameters
            let predictionMethod = options.predictionMethod;
            let extraPeriods = $.grep(mod.predictionMethods, (w) => {
                return w.value === predictionMethod;
            })[0].numPeriods;
            let level = options.level;

            //  Create an array to send back
            let datesArray = [];

            //	Get the max date info
            let lastMonth = maxDate.getMonth(),
                lastYear = maxDate.getFullYear(),
                lastDay = maxDate.getDate(),
                lastWeek = parseInt(mod.dateFormatter(maxDate,"w"));

            //  Check the level
            if (level === "days") {

                //  Figure out how many days to predict
                let endDate = new Date(lastYear,lastMonth+1,0),
                    endDay = endDate.getDate(),
                    numDays = extraPeriods ? extraPeriods : endDay - lastDay;

                //	Predict out X number of days
                for (let i = 1; i <= numDays; i++) {

                    //  Create the new date
                    let newDate = new Date(lastYear,lastMonth,lastDay+i);

                    //  Create the new data object
                    let newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            } else if (level === "weeks") {

                //  Figure out how many months to predict
                let numWeeks = extraPeriods ? extraPeriods : 52 - lastWeek;

                //  Loop through all remaining dates
                for (let i=1; i<=numWeeks; i++) {

                    //  Create the new date, add 7 days
                    let newDate = new Date(lastYear,lastMonth,lastDay);
                    newDate.setTime( newDate.getTime() + (i*7) * 86400000 );

                    //  Create the new data object
                    let newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            } else if (level === "months") {

                //  Figure out how many months to predict
                let numberOfMonths = extraPeriods ? extraPeriods : 12 - (lastMonth+1);

                //  Loop through all remaining dates
                for (let i = 1; i <= numberOfMonths; i++) {

                    //  Create the new date
                    let newDate = new Date(lastYear,lastMonth+i,1);

                    //  Create the new data object
                    let newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            } else if (level === "quarters") {

                //  Figure out how many quarters to predict
                let numQuarters = extraPeriods ? extraPeriods : 4;

                //	Loop through all remaining dates
                for (let i=1; i<=numQuarters; i++) {

                    //	Calculate new quarter month index
                    let quarterMonthIndex = (i * 3);

                    //  Create the new date
                    let newDate = new Date(lastYear,lastMonth + quarterMonthIndex,1);

                    //  Create the new data object
                    let newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            }  else if (level === "years") {

                //  Figure out how many years to predict
                let numYears = extraPeriods ? extraPeriods : 5;

                //	Loop through all remaining dates
                for (let i=1; i<=numYears; i++) {

                    //  Create the new date
                    let newDate = new Date(lastYear+i,1,1);

                    //  Create the new data object
                    let newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }
            }

            return datesArray;
        }

        //	Function to add axis labels
        function addFuturePeriodsAxisLabels(widget, chart, options, index){

            try {
                if (!options.enabled  || options.periodsAdded){
                    return;
                }

                // handle continuous time line
                if (!chart.xAxis.categories){
                    if ($$get(options,"projectedPeriods.length") > 0) {
                        let newMaxTime = options.projectedPeriods[options.projectedPeriods.length - 1][0].getTime();
                        chart.xAxis.max = newMaxTime;
                        widget.maxDate = newMaxTime;
                    }
                } else {

                    //	axis labels already added, mark flag as false
                    let chartLength = chart.xAxis.categories.length;
                    let dataLength = chart.series[index].data.length + ($$get(options, "projectedPeriods.length") || 0);
                    if (chartLength === dataLength) {
                        return;
                    }

                    //	Get the formatting mask
                    let mask = options.mask;

                    //	Loop through each point, and create a label
                    $.each(options.projectedPeriods, function () {
                        //	Parse the date
                        let thisDate = Date.parseDate ? Date.parseDate(this[0]) : this[0];

                        //	Create a formatted label
                        let label = mod.dateFormatter(thisDate, mask);

                        //	Add to the axis categories

                        chart.xAxis.categories.push(label);
                    });

                }

                //  specify that the periods have been added
                options.periodsAdded = true;

            } catch(e) {
                console.log(e);
            }
        }

        function calcRegression(currSeries, chart){

            if (currSeries.regression && !currSeries.rendered) {

                let regressionType = currSeries.regressionSettings.type,
                    regression,
                    mergedData = currSeries.data;

                // For future use - might need to calculte one regression for all series in one series
                // if ($$get(currSeries, "regressionSettings.useAllSeries")) {
                //     mergedData = [];
                //     for (let di = 0 ; di < allSeries.length ; di++) {
                //         let seriesToMerge = allSeries[di];
                //         mergedData = mergedData.concat(seriesToMerge.data);
                //     }
                // }

                try {

                    if (regressionType === "linear") {
                        regression = _linear(mergedData);
                        currSeries.type = "line";
                    } else if (regressionType === "exponential") {
                        regression = _exponential(mergedData);
                    } else if (regressionType === "logarithmic") {
                        regression = _logarithmic(mergedData);
                    }  else {
                        console.error("Invalid regression type: ", regressionType);
                        return;
                    }

                    // Add future values based on the regression equation
                    _prediction(regression, currSeries.regressionSettings.projections, regressionType);
                    currSeries.data = regression.points ;
                    currSeries.regressionOutputs = regression;
                } catch (ex){
                    let error = $$get(ex, regressionErrorKey) || errors.defaultError;

                    if (!$$get(chart, "errors.length")){
                        chart.errors = [];
                    }

                    chart.errors.push(error);
                    currSeries.visible = currSeries.showInLegend = false;
                }
            }
        }

        //  Function to predict future values
        function _prediction(data, options, regressionType) {

            //  Do we need to predict?
            if (!options.enabled) {
                //  No prediction
                return data;
            } else {
                //  Predict future values
                let equation = data.equation,
                    m = equation[0],
                    b = equation[1],
                    numPoints = data.points.length;

                //  Get the last available date
                let lastPoint = data.points[numPoints - 1];

                //  Calculate future time periods
                let forecastArray = options.projectedPeriods;

                //  Calculate future values
                for (let i = 1; i <= forecastArray.length; i++){
                    let forecastValue,
                        index = i + numPoints,
                        numOfPoints = $.extend(true, {}, lastPoint);

                    if (regressionType === "linear"){
                        //  Linear Regression: Y = m * X + b
                        forecastValue = m * index + b;
                    } else if (regressionType === "exponential") {
                        //  Exponential Regression
                        forecastValue = b * Math.pow(m, index);
                    } else if (regressionType === "logarithmic") {
                        //  Logarithmic Regression
                        forecastValue = b + (m * Math.log(index));
                    }

                    numOfPoints.y = isFinite(forecastValue) ? forecastValue : null;

                    let xVal = forecastArray[i - 1][0];
                    if (numOfPoints.x){
                        // For continuous timeline we have the x attribute set to time value
                        numOfPoints.x = $$get(xVal, "getTime") ? xVal.getTime(): xVal;
                    }

                    numOfPoints.selectionData = {0:xVal};

                    //  Add value to the array
                    data.points.push(numOfPoints);
                }

                return data;
            }
        }

        function getXYFromPoint(datapoint, indexForDateX, fixXToEpsilon = false, fixYToEpsilon = false){

            let x = indexForDateX + 1,
                y = datapoint.y;

            if (!defined(indexForDateX)){
                x =  datapoint.x || datapoint.selectionData[0];
            }

            y = fixYToEpsilon && y === 0? epsilon : y;
            x = fixXToEpsilon && x === 0? epsilon : x;

            return [x,y];
        }

        /**
         * Code extracted from https://github.com/Tom-Alexander/regression-js/
         */
        function _exponential(data) {

            let [m, b] = calcLinearEquation(data, true, true),
                results = [];

            let A = Math.exp(m);
            let R = Math.exp(b);

            for (let i = 0, len = data.length; i < len; i++) {
                let newPoint = $.extend(true, {}, data[i]);
                let pointArr = getXYFromPoint(newPoint, i);
                // y = ae ^bx
                newPoint.y = R * Math.pow(A, pointArr[0]);
                results.push(newPoint);
            }

            let string = `y =  ${Math.round(A*100) / 100} e^('${Math.round(R*100) / 100}x)`;

            return {equation: [A, R], points: results, string: string};
        }

        /**
         * Code extracted from https://github.com/Tom-Alexander/regression-js/
         * Human readable formulas:
         *
         *              N * Σ(XY) - Σ(X)
         * intercept = ---------------------
         *              N * Σ(X^2) - Σ(X)^2
         *
         * correlation = N * Σ(XY) - Σ(X) * Σ (Y) / √ (  N * Σ(X^2) - Σ(X) ) * ( N * Σ(Y^2) - Σ(Y)^2 ) ) )
         *
         */
        function _linear(data) {

            let results = [];

            let [m, b] = calcLinearEquation(data);

            for (let i = 0, len = data.length; i < len; i++) {
                let newPoint = $.extend(true, {}, data[i]),
                    pointArr = getXYFromPoint(newPoint, i);

                if (pointArr[0] !== "N/A") {
                    newPoint.y = pointArr[0] * m + b;
                    results.push(newPoint);
                }
            }

            let string = `y =  ${Math.round(m*100) / 100} x + ${Math.round(b*100) / 100}`;
            return {equation: [m, b], points: results, string: string};
        }

        /*
        * Calculate linear equation for linear, exponential and logarithmic regressions
         */
        function calcLinearEquation(data, isExponential, throwIfNullOrZero = false, isLogarithmic = false){
            let sum = [0, 0, 0, 0],
                N = 0;

            for (let n=0; n < data.length; n++) {
                let point = getXYFromPoint(data[n], n, isExponential, isExponential);

                if (isExponential && (point[1] < 0)){
                    throw {[regressionErrorKey]:errors.exponential.negative};
                } else if (!defined(point[1]) && throwIfNullOrZero){
                    throw {[regressionErrorKey]:errors.exponential.nulls};
                }

                if (isExponential){
                    point[1] = Math.log(point[1]);
                }

                if (isLogarithmic){
                    point[0] = Math.log(point[0]);
                }

                // Only take values that are not N/A
                if (point[0] !== "N/A" && defined(point[1]) && !isNaN(point[1])) {
                    sum[0] += point[0]; //Σ(X)
                    sum[1] += point[1]; //Σ(Y)
                    sum[2] += point[0] * point[0]; //Σ(XX)
                    sum[3] += point[0] * point[1]; //Σ(XY)
                    N++;
                }
            }

            /*
            * Calculate m and b for the formula:
            * y = x * m + b
            */
            let divider = (N * sum[2] - sum[0] * sum[0]);
            if (divider === 0){
                divider = epsilon;
            }

            // (N * Σ(XY) - ΣX * Σ(Y)) / (N * Σ(XX) - Σ(X) * Σ(X))
            let m = (N * sum[3] - sum[0] * sum[1]) / divider;


            let secondDivider = (m * sum[0]) / N;
            if (divider === 0){
                secondDivider = epsilon;
            }

            // (Σ(Y)/N) - (m*Σ(X))/N
            let b = (sum[1] / N) - secondDivider;

            return [m, b];
        }

        /**
         *  Code extracted from https://github.com/Tom-Alexander/regression-js/
         */
        function _logarithmic(data) {
            let [A, B] = calcLinearEquation(data, false, true, true),
                results = [];

            for (let i = 0;i < data.length; i++) {
                let newPoint = $.extend(true, {},data[i]),
                    pointArr = getXYFromPoint(newPoint, i, true, true);

                if (pointArr[0] !== "N/A") {
                    newPoint.y = B + A * Math.log(pointArr[0]);
                    results.push(newPoint);
                }
            }

            let string = `y = ${Math.round(A*100) / 100} + ${Math.round(B*100) / 100} ln(x)`;

            return {equation: [A, B], points: results, string: string};
        }

    }
]);